@@ -1,8 +1,15 @@
+use std::{fmt, mem};
+
 use thiserror::Error;
 
 use crate::tree::Kind;
 use crate::{Id, Span, Tree};
 
+/// A checkpoint which indicates a position in the tree where an element can be
+/// optionally inserted.
+#[derive(Debug, Clone, Copy)]
+pub struct Checkpoint(usize);
+
 /// Error raised by [TreeBuilder::end_node] if there currently is no node being
 /// built.
 #[derive(Debug, Error)]
@@ -23,21 +30,21 @@ pub(crate) struct Element<T> {
     pub(crate) data: T,
     /// The kind of the element.
     pub(crate) kind: Kind,
-    /// Next sibling number.
-    pub(crate) next: Option<usize>,
-    /// Head to the next child element.
-    pub(crate) child: Option<usize>,
+    /// Next sibling id.
+    pub(crate) next: usize,
+    /// The first child element.
+    pub(crate) first: usize,
 }
 
-/// A syntax tree.
+/// A syntax tree builder.
 #[derive(Debug)]
 pub struct TreeBuilder<T> {
-    ///  Data in the tree being built.
-    data: slab::Slab<Element<T>>,
+    /// Data in the tree being built.
+    pub(crate) data: Vec<Element<T>>,
     /// Nodes currently being built.
     stack: Vec<usize>,
     /// The last sibling inserted.
-    sibling: Option<usize>,
+    sibling: usize,
 }
 
 /// Build a new syntax tree.
@@ -99,7 +106,7 @@ impl<T> TreeBuilder<T> {
     pub fn start_node(&mut self, data: T) -> Id {
         let id = self.insert(data, Kind::Node);
         self.stack.push(id);
-        Id(id.try_into().expect("identifier out of bounds"))
+        Id(id)
     }
 
     /// End a node being built. This call must be balanced with
@@ -134,7 +141,7 @@ impl<T> TreeBuilder<T> {
             None => return Err(EndNodeError),
         };
 
-        self.sibling = Some(head);
+        self.sibling = head;
         Ok(())
     }
 
@@ -169,8 +176,121 @@ impl<T> TreeBuilder<T> {
     /// ```
     pub fn token(&mut self, data: T, span: Span) -> Id {
         let id = self.insert(data, Kind::Token(span));
-        self.sibling = Some(id);
-        Id(id.try_into().expect("identifier out of bounds"))
+        self.sibling = id;
+        Id(id)
+    }
+
+    /// Get a checkpoint corresponding to the current position in the tree.
+    ///
+    /// ```
+    /// use anyhow::Result;
+    /// use syntree::{print, Span, TreeBuilder};
+    ///
+    /// #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+    /// enum Syntax {
+    ///     Root,
+    ///     Number,
+    ///     Lit,
+    ///     Whitespace,
+    /// }
+    ///
+    /// # fn main() -> anyhow::Result<()> {
+    /// let mut b = TreeBuilder::new();
+    ///
+    /// let c = b.checkpoint();
+    ///
+    /// b.start_node(Syntax::Number);
+    /// b.token(Syntax::Lit, Span::new(1, 2));
+    /// b.end_node()?;
+    ///
+    /// b.token(Syntax::Whitespace, Span::new(2, 5));
+    ///
+    /// b.start_node(Syntax::Number);
+    /// b.token(Syntax::Lit, Span::new(5, 7));
+    /// b.token(Syntax::Lit, Span::new(7, 9));
+    /// b.end_node()?;
+    ///
+    /// b.insert_node_at(c, Syntax::Root);
+    ///
+    /// let tree = b.build()?;
+    ///
+    /// let root = tree.first().unwrap();
+    /// assert_eq!(*root.data(), Syntax::Root);
+    /// assert_eq!(root.children().count(), 3);
+    /// # Ok(()) }
+    /// ```
+    pub fn checkpoint(&self) -> Checkpoint {
+        Checkpoint(self.data.len())
+    }
+
+    /// Insert a node that wraps from the given checkpointed location.
+    ///
+    /// ```
+    /// use anyhow::Result;
+    /// use syntree::{print, Span, TreeBuilder};
+    ///
+    /// #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+    /// enum Syntax {
+    ///     Root,
+    ///     Number,
+    ///     Lit,
+    ///     Whitespace,
+    /// }
+    ///
+    /// # fn main() -> anyhow::Result<()> {
+    /// let mut b = TreeBuilder::new();
+    ///
+    /// let c = b.checkpoint();
+    ///
+    /// b.start_node(Syntax::Number);
+    /// b.token(Syntax::Lit, Span::new(1, 2));
+    /// b.end_node()?;
+    ///
+    /// b.token(Syntax::Whitespace, Span::new(2, 5));
+    ///
+    /// b.start_node(Syntax::Number);
+    /// b.token(Syntax::Lit, Span::new(5, 7));
+    /// b.token(Syntax::Lit, Span::new(7, 9));
+    /// b.end_node()?;
+    ///
+    /// b.insert_node_at(c, Syntax::Root);
+    ///
+    /// let tree = b.build()?;
+    ///
+    /// let root = tree.first().unwrap();
+    /// assert_eq!(*root.data(), Syntax::Root);
+    /// assert_eq!(root.children().count(), 3);
+    /// # Ok(()) }
+    /// ```
+    pub fn insert_node_at(&mut self, c: Checkpoint, data: T) -> Id
+    where
+        T: fmt::Debug,
+    {
+        // With the layout of this data structure this is a fairly simple
+        // operation.
+        let child = self.data.len();
+
+        let removed = match self.data.get_mut(c.0) {
+            Some(entry) => {
+                let new = Element {
+                    data,
+                    kind: Kind::Node,
+                    next: usize::MAX,
+                    first: child,
+                };
+
+                mem::replace(entry, new)
+            }
+            None => {
+                return Id(self.insert(data, Kind::Node));
+            }
+        };
+
+        self.data.push(removed);
+
+        // The current sibling is the newly replaced node in the tree.
+        self.sibling = c.0;
+        Id(c.0)
     }
 
     /// Construct a tree.
@@ -231,7 +351,7 @@ impl<T> TreeBuilder<T> {
     /// ```
     pub fn build(&self) -> Result<Tree<T>, IntoTreeError>
     where
-        T: Copy,
+        T: fmt::Debug + Copy,
     {
         if !self.stack.is_empty() {
             return Err(IntoTreeError);
@@ -240,29 +360,27 @@ impl<T> TreeBuilder<T> {
         Ok(crate::convert::builder_to_tree(self))
     }
 
-    /// Get the element with the given `id`.
-    pub(crate) fn get(&self, id: usize) -> Option<&Element<T>> {
-        self.data.get(id)
-    }
-
     /// Insert a new node.
     fn insert(&mut self, data: T, kind: Kind) -> usize {
-        let new = self.data.insert(Element {
+        let new = self.data.len();
+
+        let prev = std::mem::replace(&mut self.sibling, usize::MAX);
+        let parent = self.stack.last().copied().unwrap_or(usize::MAX);
+
+        self.data.push(Element {
             data,
             kind,
-            next: None,
-            child: None,
+            next: usize::MAX,
+            first: usize::MAX,
         });
 
-        if let Some(n) = self.sibling.take().and_then(|id| self.data.get_mut(id)) {
-            n.next = Some(new);
+        if let Some(prev) = self.data.get_mut(prev) {
+            prev.next = new;
         }
 
-        let last = self.stack.last().copied();
-
-        if let Some(n) = last.and_then(|id| self.data.get_mut(id)) {
-            if n.child.is_none() {
-                n.child = Some(new);
+        if let Some(n) = self.data.get_mut(parent) {
+            if n.first == usize::MAX {
+                n.first = new;
             }
         }
 
@@ -275,7 +393,7 @@ impl<T> Default for TreeBuilder<T> {
         Self {
             data: Default::default(),
             stack: Vec::new(),
-            sibling: None,
+            sibling: usize::MAX,
         }
     }
 }
